/*
**    Projet etudiants CPE Lyon
**    Copyright (C) 2012 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef MESH_HPP
#define MESH_HPP

#include <v3.hpp>

#include <iostream>
#include <vector>
#include <map>
#include <list>

/** \brief Namespace for cpe classes */
namespace cpe
{
class matrix3;
}

namespace cpe
{

    /** \brief basic mesh class


    */
    class mesh
    {

    public:

        // ********************************************* //
        // ********************************************* //
        //  CONSTRUCTORS
        // ********************************************* //
        // ********************************************* //

        /** \brief empty constructor */
        mesh();
        /** \brief constructor from std::pair <vertex,connectivity> */
        mesh(const std::pair <std::vector<double>,std::vector<unsigned int> >& input);


        // ********************************************* //
        // ********************************************* //
        //  Get value
        // ********************************************* //
        // ********************************************* //

        /** \brief get vertex k */
        v3 get_vertex(const unsigned int& k_vertex) const;
        /** \brief get k_v (k_v in [0,2]) vertex of triangle k_tri */
        v3 get_vertex(const unsigned int& k_tri,const unsigned int& k_v) const;
        /** \brief get connectivity polygon (to do)*/
        std::vector<unsigned int> get_connectivity(const unsigned int& k_polygon) const;

        /** \brief set a 3D vertex */
        void set_vertex(const unsigned int& k_vertex,const v3& new_vertex);
        /** \brief set connectivity */
        void set_connectivity(const unsigned int& k_polygon,const std::vector<unsigned int>& new_triangle);



        /** \brief set texture for vertex k_vertex */
        void set_texture(const unsigned int& k_vertex,const double& tx,const double& ty);
        /** \brief get texture tx for vertex k_vertex */
        double get_texture_tx(const unsigned int& k_vertex) const;
        /** \brief get texture ty for vertex k_vertex */
        double get_texture_ty(const unsigned int& k_vertex) const;


        // ********************************************* //
        // ********************************************* //
        //  Build shape
        // ********************************************* //
        // ********************************************* //

        /** \brief add a vertex in the mesh */
        void add_vertex(const v3& x);
        /** \brief add a vertex in the mesh */
        void add_vertex(const double& x,const double& y,const double& z);
        /** \brief add indices of a triangle in the mesh */
        void add_triangle(const unsigned int& u0,const unsigned int& u1,const unsigned int& u2);


        // ********************************************* //
        // ********************************************* //
        //  Size
        // ********************************************* //
        // ********************************************* //

        /** \brief number of vertices */
        unsigned int vertex_number() const;
        /** \brief number of triangles */
        unsigned int triangle_number() const;

        // ********************************************* //
        // ********************************************* //
        //  Connectivity information
        // ********************************************* //
        // ********************************************* //

        /** \brief build one ring */
        std::map<int,std::list<int> > ring() const;
        /** \brief build one star */
        std::map<int,std::list<int> > star() const;


        // ********************************************* //
        // ********************************************* //
        //  Normal
        // ********************************************* //
        // ********************************************* //

        /** \brief compute per vertex normal */
        std::vector <double> normal_vertex() const;
        /** \brief compute per polygon normal */
        std::vector <double> normal_polygon() const;


        /** \brief swap triangle order (inverse normal direction) */
        void inverse_triangle_order();


        // ********************************************* //
        // ********************************************* //
        //  INTERNAL ACCESS
        // ********************************************* //
        // ********************************************* //

        /** \brief get internal vertex */
        const std::vector<double>& get_vertex() const;
        /** \brief get internal vertex */
        std::vector<double>& get_vertex();

        /** \brief get internal connectivity */
        const std::vector <unsigned int>& get_connectivity() const;
        /** \brief get internal connectivity */
        std::vector <unsigned int>& get_connectivity();

        /** \brief get the internal vector of texture */
        const std::vector<double>& get_texture() const;
        /** \brief get the internal vector of texture */
        std::vector<double>& get_texture();

        // ********************************************* //
        // ********************************************* //
        //  Assert
        // ********************************************* //
        // ********************************************* //

        /** \brief check vertex index are within the bounds */
        void assert_vertex(const unsigned int& k_vertex) const;

        /** \brief check polygon index are within the bounds */
        void assert_connectivity(const unsigned int& k_connectivity) const;


        // ********************************************* //
        // ********************************************* //
        //  IN/OUT
        // ********************************************* //
        // ********************************************* //

        /** \brief load off file */
        static mesh load_off_file(const std::string& filename);
        /** \brief save mesh as off file */
        void save_off_file(const std::string& filename) const;


        // ********************************************* //
        // ********************************************* //
        //  Deformation
        // ********************************************* //
        // ********************************************* //

        /** \brief scale and translate the mesh to fit into the unit cube */
        void scale_unity();

        /** \brief scale the mesh */
        void scale(const double& value);
        /** \brief matrix multiplication */
        void apply(const matrix3& M);
        /** \brief translate */
        void translate(const v3& tr);




    private:

        /** \brief vertex vector (size=3xN_vertex) */
        std::vector <double> vertex;
        /** \brief connectivity vector (only triangles) (size=3xN_poly)*/
        std::vector <unsigned int> connectivity;

        /** \brief texture vector (size=2xN_vertex)
            \note Assume texture connectivity=vertex connectivity
         */
        std::vector <double> texture;

    };
}

#endif
