/*
**    Projet etudiants CPE Lyon
**    Copyright (C) 2012 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef STRING_HELPER_H
#define STRING_HELPER_H

#include <stdlib.h>
#include <vector>
#include <iostream>
#include <sstream>
#include <string>



namespace cpe
{



  /// A helper class to manipulate string (token, convert, ...)
  /** A class to convert a string into value or value into string */

  class string_converter
  {


  public:


      /** \brief return the value of the given string
          \return the value asked and a boolean is set to 0(!=0) if the conversion succed (failed)

          example: double val=mesh_conv::string_converter::value_of<double> ("4.54");
      */
      template <typename T>
              static T value_of(const std::string& in,bool *is_ok=0)
      {
          T obj;
          std::istringstream is(in);
          bool _is_ok=bool(is>>obj);
          if(is_ok!=0)
              *is_ok=_is_ok;
          return obj;
      }

      /** \brief return the value of the given vector of string */
      template <typename T>
              static std::vector <T> value_of(const std::vector <std::string>& in,bool *is_ok=0)
      {

          if(is_ok!=0)
              *is_ok=true;

          int N=in.size();
          std::vector <T> obj(N);
          bool temp_is_ok=true;
          for(int k=0;k<N;++k)
          {
              obj[k]=value_of<T> (in[k],&temp_is_ok);
              if(is_ok!=0 && temp_is_ok==false)
                  *is_ok=false;
          }
          return obj;
      }

    /** \brief get the string corresponding to the given value

        example: std::string val=mesh_conv::string_converter::to_string(4.0*78.5);
    */
    template <typename T>
    static std::string to_string(const T& t)
    {
      std::ostringstream oss;
      oss<<t;
      return std::string(oss.str());
    }

    /** \brief fill the appropriate number of zero to have the same size

        exemple zero_padding("62",5)="00062"

        exemple zero_padding("c;7",4)="0x;7"
    */
    static std::string zero_padding(const std::string& input,const int& zero_number=4);

    /** \brief helper to write padded string from a number easily */
    template <typename T>
            static std::string to_string_padded(const T& x,const unsigned int& zero_number=4)
    {return string_converter::zero_padding(string_converter::to_string(x),zero_number);}

    /** \brief return the lower case of a string */
    static std::string to_lower(const std::string& input);
    /** \brief return the upper case of a string */
    static std::string to_upper(const std::string& input);

    /** \brief delete the empty space of a vector of string */
    static std::vector <std::string> delete_empty(const std::vector <std::string>& in);


    /** \brief extract the last number part of a string and give the number of zeros
        \param int: number found (-1 if failed)
        \param int: number of zeros (-1 if failed)
        \param string: first part of the word
        \param string: last part of the word

        ex. my_file_001.dat -> number=1, zeros=2, part_1="my_file_" part_2=".dat"
    */
    static std::pair<std::pair<int,int>,std::pair<std::string,std::string> > extract_number_part(const std::string& filename);

    /** \brief load file sequence

        Try to load every filename following the numbers on disk
    */
    static std::vector<std::string> load_filename_sequence(const std::string filename,const unsigned int& iteration=1);


  private:
  };


 /// Helper class to tokenize easily a string
  class string_tokenizer
  {

  public:
      /** \brief tokenize a given string
          \return the vector of string tokenized
          \param str the input string
          \param delimiters the delimiter between the string ("," ";" " " ...), default is empty space

          exemple std::vector <std::string> mesh_conv::string_tokenizer.tokenize(std::string("4.5,87.1,10"),",");
      */
      static std::vector <std::string> tokenize(const std::string& str,const std::string& delimiters=" ");
  private:
  };

  /** \brief helper class with files */
  class file_helper
  {
      public:

      /** \brief copy the content of a file into an other one
          \return an instance of file_helper
          \param std::string& the name of the file to copy
          \param std::string& the name of the file receiving the copy
      */
      static file_helper copy(const std::string& input_filename,const std::string& output_filename);

      private:
  };


}

#endif
