/*
**    Projet etudiants CPE Lyon
**    Copyright (C) 2012 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <quaternion.hpp>

#include <v3.hpp>
#include <matrix3.hpp>
#include <cmath>

namespace cpe
{

    quaternion::quaternion():v4(1,0,0,0){}
    quaternion::quaternion(const v4& entries):v4(entries){}
    quaternion::quaternion(const v3& axis,const double& angle)
    {
        double sin_phi=std::sin(angle/2.0);
        double cos_phi=std::cos(angle/2.0);
        v3 n_axis=axis.normalized();

        x() = n_axis[0]*sin_phi;
        y() = n_axis[1]*sin_phi;
        z() = n_axis[2]*sin_phi;
        w() =           cos_phi;
    }

    v3 quaternion::axis() const {return v3(x(),y(),z()).normalized();}
    double quaternion::angle() const {return 2*std::acos(w());}
    matrix3 quaternion::matrix() const
    {
        double x2=x()*x();
        double y2=y()*y();
        double z2=z()*z();
        double xy=x()*y();
        double xz=x()*z();
        double yz=y()*z();
        double wx=w()*x();
        double wy=w()*y();
        double wz=w()*z();

        return matrix3(1-2*(y2+z2),  2*(xy-wz),  2*(xz+wy),
                       2*(xy+wz),1-2*(x2+z2),  2*(yz-wx),
                       2*(xz-wy),  2*(yz+wx),1-2*(x2+y2));
    }
    quaternion quaternion::conjugated() const
    {
        return quaternion(v4(-x(),-y(),-z(),w()));
    }
    quaternion quaternion::slerp(const quaternion& q0,const quaternion& q1,const double& t)
    {
        double cos_theta=q0.dot(q1);
        double theta=std::acos(cos_theta);
        double sin_theta=std::sqrt(1-cos_theta*cos_theta);

        double epsilon=1e-6;
        if(std::fabs(sin_theta)<epsilon) {return (1-t)*q0+t*q1;}

        return (sin((1-t)*theta)*q0 + sin(t*theta)*q1) / sin_theta;
    }

    quaternion operator*(const quaternion& q0,const quaternion& q1)
    {
        v3 v0(q0.x(),q0.y(),q0.z());
        v3 v1(q1.x(),q1.y(),q1.z());

        const double& w0=q0.w();
        const double& w1=q1.w();

        return quaternion(v4(w0*v1+w1*v0-v0.cross(v1),w0*w1-v0.dot(v1)));

    }

    quaternion& quaternion::operator*=(const quaternion& q1)
    {
        const double& w0=w();
        const double& w1=q1.w();

        v3 v0=v3(x(),y(),z());
        v3 v1=v3(q1.x(),q1.y(),q1.z());

        v3 a=w0*v1+w1*v0-v0.cross(v1);
        w()=w0*w1-v0.dot(v1);
        x()=a.x();
        y()=a.y();
        z()=a.z();

        return *this;

    }

    v3 operator*(const quaternion& q,const v3& v)
    {
        double x2=q.x()*q.x();
        double y2=q.y()*q.y();
        double z2=q.z()*q.z();
        double xy=q.x()*q.y();
        double xz=q.x()*q.z();
        double yz=q.y()*q.z();
        double wx=q.w()*q.x();
        double wy=q.w()*q.y();
        double wz=q.w()*q.z();

        return v3( (1-2*(y2+z2))*v.x() + 2*(xy-wz)*v.y() + 2*(xz+wy)*v.z(),
                   2*(xy+wz)*v.x()+(1-2*(x2+z2))*v.y()+2*(yz-wx)*v.z(),
                   2*(xz-wy)*v.x()+2*(yz+wx)*v.y()+(1-2*(x2+y2))*v.z() );

    }

}
