/*
**    Projet etudiants CPE Lyon
**    Copyright (C) 2012 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <glwidget.hpp>

#include <matrix3.hpp>

#include <Qt/QtOpenGL>
#include <QtOpenGL/QGLWidget>
#include <window.hpp>



//fonction d'initialisation des variables
glwidget::glwidget(QWidget *parent)
  : QGLWidget(parent)
{
    QWidget::setFocusPolicy(Qt::StrongFocus);

    print_help_start();
}

//destructeur
glwidget::~glwidget()
{
}

//affichage de l'aide textuelle de navigation en debut de programme
void glwidget::print_help_start() const
{
    std::cout<<"**Hello Visualiseur**\n"<<std::endl;
    std::cout<<"===================="<<std::endl;
    std::cout<<"Pour se deplacer: "<<std::endl;
    std::cout<<"      Click gauche => rotation "<<std::endl;
    std::cout<<"      Click droit  => zoom"<<std::endl;
    std::cout<<"SHIFT+Click gauche => translation gauche/droite, haut/bas "<<std::endl;
    std::cout<<"SHIFT+Click gauche => Avant/arriere "<<std::endl;
    std::cout<<"====================\n"<<std::endl;
}

//fonction d'affichage principale: appellee en permanence
void glwidget::paintGL()
{

    //Viewport = dimensions de l'ecran d'affichage
    glViewport (0, 0, nav.screen_size_x(),nav.screen_size_y());
    //Effacement des donnees a l'ecran
    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    //Matrice de Projection (camera)
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity ();
    std::vector<double> proj_param=nav.projection_param();
    glFrustum(proj_param[0], -proj_param[0], proj_param[1], -proj_param[1], proj_param[2], proj_param[3]);


    //Matrice de Modelview (matrice du monde)
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glTranslated(0,0,static_cast<GLdouble>(nav.dist_cam())); //zoom
    std::vector<double> m=nav.camera_orientation().to_4d();
    glMultMatrixd(&m[0]); //rotation
    cpe::v3 tr=nav.camera_translation();  //translation
    glTranslated(static_cast<GLdouble>(tr[0]),static_cast<GLdouble>(tr[1]),static_cast<GLdouble>(tr[2]));


    //Affichage de la scene
    current_scene.draw_scene();


    //force affichage
    glFlush ();

}

//fonction d'initialisation d'OpenGL executee une unique fois au debut du programme
void glwidget::initializeGL()
{
    //on efface en blanc
    qglClearColor(QColor(255,255,255));

    //on veut un affichage lisse
    glShadeModel(GL_SMOOTH);
    glEnable(GL_LINE_SMOOTH);
    glHint(GL_LINE_SMOOTH_HINT,GL_NICEST);

    //on active le buffer de profondeur
    glEnable(GL_DEPTH_TEST);

    //mise en place de la lumiere
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    glFrontFace( GL_CW ) ;
    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    float light_specular[4]={0.01f,0.01f,0.01f,1.0f};
    glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
    float light_diffuse[4]={0.8f,0.8f,0.8f,1.0f};
    glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
    float material_specular [4] = {1.0f,1.0f,1.0f,1.0f};
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, material_specular);

    //fonction de melange de couleur
    glEnable(GL_BLEND);
    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable(GL_COLOR_MATERIAL);


    try
    {
        current_scene=scene(this);

        //chargement de la scene au lancement du programme
        current_scene.load_model();
    }
    catch(std::string e)
    {
        std::cout<<"************************************************"<<std::endl;
        std::cout<<"Exception: "<<e<<std::endl;
        std::cout<<"************************************************"<<std::endl;
        exit(1);
    }
}



//redimensionnement de la fenetre
void glwidget::resizeGL(int width, int height)
{
    nav.screen_size_x()=width;nav.screen_size_y()=height;glViewport(0,0, width, height);
}

//click souris
void glwidget::mousePressEvent(QMouseEvent *event)
{
    nav.x_previous()=event->x();
    nav.y_previous()=event->y();


    updateGL();
}

//relachement de la souris
void glwidget::mouseReleaseEvent(QMouseEvent *event)
{
    updateGL();
}

//lorsque la souris se deplace
void glwidget::mouseMoveEvent(QMouseEvent *event)
{
    int x=event->x();
    int y=event->y();

    //gestion d'une trackball = rotation de la camera
    int shift_pressed = (event->modifiers() & Qt::ShiftModifier);
    if (!shift_pressed && (event->buttons() & Qt::LeftButton) )
        nav.trackball_rotate(x,y);
    else if (!shift_pressed && (event->buttons() & Qt::RightButton))
        nav.trackball_zoom(y);

    double dL=0.0001*(1+10*abs(nav.dist_cam()));
    if( shift_pressed && (event->buttons() & Qt::LeftButton) )
    {
        nav.go_up(dL*(y-nav.y_previous()));
        nav.go_right(-dL*(x-nav.x_previous()));
    }
    if( shift_pressed && (event->buttons() & Qt::RightButton) )
        nav.go_forward(5*dL*(y-nav.y_previous()));


    nav.x_previous()=x;
    nav.y_previous()=y;

    updateGL();

}

//lorsque l'on appuis sur une touche
void glwidget::keyPressEvent(QKeyEvent *event)
{
    QString k=event->text();


    int current=event->key();
    Qt::KeyboardModifiers mod=event->modifiers();

    //si on appuis sur Q (majuscule), on quitte le programme
    if( (mod&Qt::ShiftModifier)!=0 && (current==Qt::Key_Q) )
    {
        std::cout<<"\n[EXIT OK]\n\n"<<std::endl;
        this->window()->close();
    }


    QGLWidget::keyPressEvent(event);
    updateGL();
}


